/*
 * A C++ scanner that uses the longest match scanning method. This example
 * differs from other examples of scanning. Each run of the state machine
 * matches one token. This method results in a smaller state machine since the
 * final kleene star is omitted and therefore every state does not need to get
 * all the transitions of the start state.
 *
 * << <= <<= >> >= >>= are left out since angle brackets are used in templates.
 *
 * In this example the the finishing action operator "@" is used to remember
 * that a token was matched every time the machine enters into a final state.
 * After the machine moves into the error state we take the last matching
 * token. If two tokens match the same input, the token referenced later in the
 * machine construction will have it's actions executed second (Ragel ensures
 * this) and will overwrite any previous token. 
 */

#include <iostream>
using namespace std;

#define TK_Dlit 256
#define TK_Slit 257
#define TK_Float 258
#define TK_Id 259
#define TK_NameSep 260
#define TK_Arrow 261
#define TK_PlusPlus 262
#define TK_MinusMinus 263
#define TK_ArrowStar 264
#define TK_DotStar 265
#define TK_ShiftLeft 266
#define TK_ShiftRight 267
#define TK_IntegerDecimal 268
#define TK_IntegerOctal 269
#define TK_IntegerHex 270
#define TK_EqualsEquals 271
#define TK_NotEquals 272
#define TK_AndAnd 273
#define TK_OrOr 274
#define TK_MultAssign 275
#define TK_DivAssign 276
#define TK_PercentAssign 277
#define TK_PlusAssign 278
#define TK_MinusAssign 279
#define TK_AmpAssign 280
#define TK_CaretAssign 281
#define TK_BarAssign 282
#define TK_DotDotDot 283
#define TK_Whitespace 284
#define TK_Comment 285

#define BUFSIZE 2048

int tok;
char buf[BUFSIZE], *bfront, *bend, *tokend;

%% Scanner
	struct {};
	init { 
		tok = 0;
		tokend = 0;
	}

	# Single and double literals.
	slit = ( 'L'? "'" ( [^'\\\n] | /\\./ )* "'" ) @{tok = TK_Slit;};
	dlit = ( 'L'? '"' ( [^"\\\n] | /\\./ )* '"' ) @{tok = TK_Dlit;};

	# Identifiers
	id = ( [a-zA-Z_] [a-zA-Z0-9_]* ) @{tok = TK_Id;};

	# Floating literals.
	fract_const = digit* '.' digit+ | digit+ '.';
	exponent = [eE] [+\-]? digit+;
	float_suffix = [flFL];
	float = 
		( fract_const exponent? float_suffix? |
		digit+ exponent float_suffix? ) @{tok = TK_Float;};
	
	# Integer decimal. Leading part buffered by float.
	integer_decimal = ( ( '0' | [1-9] [0-9]* ) [ulUL]{0,3} ) @{tok = TK_IntegerDecimal;};

	# Integer octal. Leading part buffered by float.
	integer_octal = ( '0' [0-9]+ [ulUL]{0,2} ) @{tok = TK_IntegerOctal;};

	# Integer hex. Leading 0 buffered by float.
	integer_hex = ( '0' ( 'x' [0-9a-fA-F]+ [ulUL]{0,2} ) ) @{tok = TK_IntegerHex;};

	# Only buffer the second item, first buffered by symbol. */
	namesep = '::' @{tok = TK_NameSep;};
	deqs = '==' @{tok = TK_EqualsEquals;};
	neqs = '!=' @{tok = TK_NotEquals;};
	and_and = '&&' @{tok = TK_AndAnd;};
	or_or = '||' @{tok = TK_OrOr;};
	mult_assign = '*=' @{tok = TK_MultAssign;};
	percent_assign = '%=' @{tok = TK_PercentAssign;};
	plus_assign = '+=' @{tok = TK_PlusAssign;};
	minus_assign = '-=' @{tok = TK_MinusAssign;};
	amp_assign = '&=' @{tok = TK_AmpAssign;};
	caret_assign = '^=' @{tok = TK_CaretAssign;};
	bar_assign = '|=' @{tok = TK_BarAssign;};
	plus_plus = '++' @{tok = TK_PlusPlus;};
	minus_minus = '--' @{tok = TK_MinusMinus;};
	arrow = '->' @{tok = TK_Arrow;};
	arrow_star = '->*' @{tok = TK_ArrowStar;};
	dot_star = '.*' @{tok = TK_DotStar;};

	# Buffer both items. *
	div_assign = '/=' @{tok = TK_DivAssign;};

	# Three char compounds, first item already buffered. */
	dot_dot_dot = '...' @{tok = TK_DotDotDot;};

	# All compunds
	compound = namesep | deqs | neqs | and_and | or_or | mult_assign |
			div_assign | percent_assign | plus_assign | minus_assign |
			amp_assign | caret_assign | bar_assign | plus_plus | minus_minus |
			arrow | arrow_star | dot_star | dot_dot_dot;

	# Single char symbols.
	symbol = ( punct - [_"'] ) @{tok = fc;};

	# Comments and whitespace.
	commc = '/*' ( any* $0 '*/' @1 ) @{tok = TK_Comment;};
	commcc = '//' ( any* $0 '\n' @1 ) @{tok = TK_Comment;};
	whitespace = ( -128..32 | 127 )+ @{tok = TK_Whitespace;};

	# All outside code tokens.
	tokens = ( 
		id | slit | dlit | float | integer_decimal | 
		integer_octal | integer_hex | compound | symbol |
		commc | commcc | whitespace );

	main := tokens @{tokend=fpc;};
%%

int main()
{
	Scanner scanner;

	/* Do the first read. */
	bfront = buf;
	cin.read( buf, BUFSIZE );
	bend = buf + cin.gcount();

	while ( bfront < bend || !cin.eof() ) {
		scanner.init();
		int len = bend - bfront;
		int rtn = scanner.execute( bfront, len );
		/* If we don't have a token yet, get some more data. */
		if ( rtn >= 0 && ! cin.eof() ) {
			/* Maybe shift data to the left. */
			if ( len > 0 && bfront > buf ) {
				memmove( buf, bfront, bend-bfront );
				if ( tokend != 0 )
					tokend -= bfront - buf;
			}

			/* Read in some more data. */
			bfront = buf;
			bend = buf+len;
			cin.read( bend, BUFSIZE-len );
			bend = buf + len + cin.gcount();

			/* Execute on any new data read. */
			char *mfront = bfront+len;
			rtn = scanner.execute( bfront+len, bend-mfront );

			/* If still don't have a token, then bail claiming not enough
			 * space. */
			if ( rtn >= 0 && !cin.eof() ) {
				cerr << "TOKEN TOO BIG" << endl;
				exit(1);
			}
		}

		if ( tok == 0 ) {
			cout.flush();
			cerr << "PARSE ERROR" << endl;
			exit(1);
		}
		else {
			/* Now process token. */
			len = tokend - bfront + 1;

			if ( tok != TK_Whitespace && tok != TK_Comment )
				cout << '<' << tok << '>';
			for ( int i = 0; i < len; i++ )
				cout << bfront[i];
			bfront += len;
		}
	}

	return 0;
}
