#include <iostream>
using namespace std;

#define TK_Dlit 192
#define TK_Slit 193
#define TK_Float 194
#define TK_Id 195
#define TK_NameSep 197
#define TK_Arrow 211
#define TK_PlusPlus 212
#define TK_MinusMinus 213
#define TK_ArrowStar 214
#define TK_DotStar 215
#define TK_ShiftLeft 216
#define TK_ShiftRight 217
#define TK_IntegerDecimal 218
#define TK_IntegerOctal 219
#define TK_IntegerHex 220
#define TK_EqualsEquals 223
#define TK_NotEquals 224
#define TK_AndAnd 225
#define TK_OrOr 226
#define TK_MultAssign 227
#define TK_DivAssign 228
#define TK_PercentAssign 229
#define TK_PlusAssign 230
#define TK_MinusAssign 231
#define TK_AmpAssign 232
#define TK_CaretAssign 233
#define TK_BarAssign 234
#define TK_DotDotDot 240
#define TK_Whitespace 241
#define TK_Comment 242

#define BUFSIZE 1024
int tok;
char buf[BUFSIZE], *bfront, *bend, *tokend;

%% Scanner
	struct {};
	init { 
		tok = 0;
		tokend = 0;
	}

	# Single and double literals.
	slit = ( 'L'? "'" ( [^'\\\n] | /\\./ )* "'" ) @{tok = TK_Slit;};
	dlit = ( 'L'? '"' ( [^"\\\n] | /\\./ )* '"' ) @{tok = TK_Dlit;};

	# Identifiers
	id = ( [a-zA-Z_] [a-zA-Z0-9_]* ) @{tok = TK_Id;};

	# Floating literals.
	fract_const = digit* '.' digit+ | digit+ '.';
	exponent = [eE] [+\-]? digit+;
	float_suffix = [flFL];
	float = 
		( fract_const exponent? float_suffix? |
		digit+ exponent float_suffix? ) @{tok = TK_Float;};
	
	# Integer decimal. Leading part buffered by float.
	integer_decimal = ( ( '0' | [1-9] [0-9]* ) [ulUL]{0,3} ) @{tok = TK_IntegerDecimal;};

	# Integer octal. Leading part buffered by float.
	integer_octal = ( '0' [0-9]+ [ulUL]{0,2} ) @{tok = TK_IntegerOctal;};

	# Integer hex. Leading 0 buffered by float.
	integer_hex = ( '0' ( 'x' [0-9a-fA-F]+ [ulUL]{0,2} ) ) @{tok = TK_IntegerHex;};

	# Only buffer the second item, first buffered by symbol. */
	namesep = '::' @{tok = TK_NameSep;};
	deqs = '==' @{tok = TK_EqualsEquals;};
	neqs = '!=' @{tok = TK_NotEquals;};
	and_and = '&&' @{tok = TK_AndAnd;};
	or_or = '||' @{tok = TK_OrOr;};
	mult_assign = '*=' @{tok = TK_MultAssign;};
	percent_assign = '%=' @{tok = TK_PercentAssign;};
	plus_assign = '+=' @{tok = TK_PlusAssign;};
	minus_assign = '-=' @{tok = TK_MinusAssign;};
	amp_assign = '&=' @{tok = TK_AmpAssign;};
	caret_assign = '^=' @{tok = TK_CaretAssign;};
	bar_assign = '|=' @{tok = TK_BarAssign;};
	plus_plus = '++' @{tok = TK_PlusPlus;};
	minus_minus = '--' @{tok = TK_MinusMinus;};
	arrow = '->' @{tok = TK_Arrow;};
	arrow_star = '->*' @{tok = TK_ArrowStar;};
	dot_star = '.*' @{tok = TK_DotStar;};

	# Buffer both items. *
	div_assign = '/=' @{tok = TK_DivAssign;};

	# Three char compounds, first item already buffered. */
	dot_dot_dot = '...' @{tok = TK_DotDotDot;};

	# All compunds
	compound = namesep | deqs | neqs | and_and | or_or | mult_assign |
			div_assign | percent_assign | plus_assign | minus_assign |
			amp_assign | caret_assign | bar_assign | plus_plus | minus_minus |
			arrow | arrow_star | dot_star | dot_dot_dot;

	# Single char symbols.
	symbol = ( punct - [_"'] ) @{tok = fc;};

	# Comments and whitespace.
	commc = '/*' ( any* $0 '*/' @1 ) @{tok = TK_Comment;};
	commcc = '//' ( any* $0 '\n' @1 ) @{tok = TK_Comment;};
	whitespace = ( -128..32 | 127 )+ @{tok = TK_Whitespace;};

	# All outside code tokens.
	tokens = ( 
		id | slit | dlit | float | integer_decimal | 
		integer_octal | integer_hex | compound | symbol |
		commc | commcc | whitespace );

	action success {
		nonTokBuf.append(0);
		cout << nonTokBuf.data;
		cout.flush();
	}

	position = (
		'\n' @{ line += 1; col = 1; } |
		[^\n] @{ col += 1; } )*;

	main := tokens @{tokend=fpc;};
%%

int main()
{
	Scanner scanner;

	/* Do the first read. */
	bfront = buf;
	cin.read( buf, BUFSIZE );
	bend = buf + cin.gcount();

	while ( bfront < bend || !cin.eof() ) {
		scanner.init();
		int len = bend - bfront;
		int rtn = scanner.execute( bfront, len );
		/* If we don't have a token yet, get some more data. */
		if ( rtn >= 0 && ! cin.eof() ) {
			/* Maybe shift data to the left. */
			if ( len > 0 && bfront > buf ) {
				memmove( buf, bfront, bend-bfront );
				if ( tokend != 0 )
					tokend -= bfront - buf;
			}

			/* Read in some more data. */
			bfront = buf;
			bend = buf+len;
			cin.read( bend, BUFSIZE-len );
			bend = buf + len + cin.gcount();

			/* Execute on any new data read. */
			char *mfront = bfront+len;
			rtn = scanner.execute( bfront+len, bend-mfront );

			/* If still don't have a token, then bail claiming not enough
			 * space. */
			if ( rtn >= 0 && !cin.eof() ) {
				cerr << "TOKEN TOO BIG" << endl;
				exit(1);
			}
		}

		if ( tok == 0 ) {
			cout.flush();
			cerr << "PARSE ERROR" << endl;
			exit(1);
		}
		else {
			/* Now process token. */
			len = tokend - bfront + 1;

			if ( tok != TK_Whitespace && tok != TK_Comment )
				cout << '<' << tok << '>';
			for ( int i = 0; i < len; i++ )
				cout << bfront[i];
			bfront += len;
		}
	}

	return 0;
}
